// This code was autogenerated from xbus/xbus.proto, do not edit.
package api

import (
	"context"
	"log"
	"time"

	"github.com/golang/protobuf/proto"
	nats "github.com/nats-io/go-nats"
	github_com_nats_rpc_gogo_nrpc "github.com/nats-rpc/gogo-nrpc"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/nats-rpc/gogo-nrpc"
)

// DirectorServer is the interface that providers of the service
// Director should implement.
type DirectorServer interface {
	ProcessingEnd(ctx context.Context, actorID string, req ActorProcessingState) (err error)
	Log(ctx context.Context, actorID string, req ActorLogRequest) (err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForDirector = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Director",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForDirector = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Director",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForDirector = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "Director",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForDirector = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "Director",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// DirectorHandler provides a NATS subscription handler that can serve a
// subscription using a given DirectorServer implementation.
type DirectorHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  DirectorServer
}

func NewDirectorHandler(ctx context.Context, nc nrpc.NatsConn, s DirectorServer) *DirectorHandler {
	return &DirectorHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewDirectorConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s DirectorServer) *DirectorHandler {
	return &DirectorHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *DirectorHandler) Subject() string {
	return "xbus.default.director.>"
}

func (h *DirectorHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.default", 0, "director", 0, msg.Subject)
	if err != nil {
		log.Printf("DirectorHanlder: Director subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "processingend":
		var mtParams []string
		mtParams, request.Encoding, err = nrpc.ParseSubjectTail(1, request.SubjectTail)
		if err != nil {
			log.Printf("ProcessingEndHanlder: ProcessingEnd subject parsing failed: %v", err)
			break
		}
		var req ActorProcessingState
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("ProcessingEndHandler: ProcessingEnd request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForDirector.WithLabelValues(
				"ProcessingEnd", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				var innerResp nrpc.Void
				err := h.server.ProcessingEnd(ctx, mtParams[0], req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "log":
		var mtParams []string
		mtParams, request.Encoding, err = nrpc.ParseSubjectTail(1, request.SubjectTail)
		if err != nil {
			log.Printf("LogHanlder: Log subject parsing failed: %v", err)
			break
		}
		var req ActorLogRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("LogHandler: Log request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForDirector.WithLabelValues(
				"Log", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				var innerResp nrpc.Void
				err := h.server.Log(ctx, mtParams[0], req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	default:
		log.Printf("DirectorHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForDirector.WithLabelValues(
			"Director", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForDirector.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForDirector.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForDirector.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForDirector.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("DirectorHandler: Director handler failed to publish the response: %s", err)
			serverRequestsForDirector.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForDirector.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type DirectorClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewDirectorClient(nc nrpc.NatsConn) *DirectorClient {
	return &DirectorClient{
		nc:      nc,
		PkgSubject: "xbus.default",
		Subject: "director",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *DirectorClient) ProcessingEnd(actorID string, req ActorProcessingState) (err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "processingend" + "." + actorID

	// call
	var resp github_com_nats_rpc_gogo_nrpc.Void
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForDirector.WithLabelValues(
			"ProcessingEnd", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForDirector.WithLabelValues("ProcessingEnd").Observe(elapsed)
	clientCallsForDirector.WithLabelValues(
		"ProcessingEnd", c.Encoding, "success").Inc()

	return
}

func (c *DirectorClient) Log(actorID string, req ActorLogRequest) (err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "log" + "." + actorID

	// call
	var resp github_com_nats_rpc_gogo_nrpc.Void
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForDirector.WithLabelValues(
			"Log", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForDirector.WithLabelValues("Log").Observe(elapsed)
	clientCallsForDirector.WithLabelValues(
		"Log", c.Encoding, "success").Inc()

	return
}

// EnvelopeStorageServer is the interface that providers of the service
// EnvelopeStorage should implement.
type EnvelopeStorageServer interface {
	Retrieve(ctx context.Context, clientKind string, clientID string, req EnvelopePosition, pushRep func(EnvelopeFragmentAndPosition)) (err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForEnvelopeStorage = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "EnvelopeStorage",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForEnvelopeStorage = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "EnvelopeStorage",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForEnvelopeStorage = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "EnvelopeStorage",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForEnvelopeStorage = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "EnvelopeStorage",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// EnvelopeStorageHandler provides a NATS subscription handler that can serve a
// subscription using a given EnvelopeStorageServer implementation.
type EnvelopeStorageHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  EnvelopeStorageServer
}

func NewEnvelopeStorageHandler(ctx context.Context, nc nrpc.NatsConn, s EnvelopeStorageServer) *EnvelopeStorageHandler {
	return &EnvelopeStorageHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewEnvelopeStorageConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s EnvelopeStorageServer) *EnvelopeStorageHandler {
	return &EnvelopeStorageHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *EnvelopeStorageHandler) Subject() string {
	return "xbus.default.envelopestorage.>"
}

func (h *EnvelopeStorageHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.default", 0, "envelopestorage", 0, msg.Subject)
	if err != nil {
		log.Printf("EnvelopeStorageHanlder: EnvelopeStorage subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "retrieve":
		var mtParams []string
		mtParams, request.Encoding, err = nrpc.ParseSubjectTail(2, request.SubjectTail)
		if err != nil {
			log.Printf("RetrieveHanlder: Retrieve subject parsing failed: %v", err)
			break
		}
		var req EnvelopePosition
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("RetrieveHandler: Retrieve request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForEnvelopeStorage.WithLabelValues(
				"Retrieve", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.EnableStreamedReply()
			request.Handler = func(ctx context.Context)(proto.Message, error){
				err := h.server.Retrieve(ctx, mtParams[0], mtParams[1], req, func(rep EnvelopeFragmentAndPosition){
					request.SendStreamReply(&rep)
				})
				return nil, err
			}
		}
	default:
		log.Printf("EnvelopeStorageHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForEnvelopeStorage.WithLabelValues(
			"EnvelopeStorage", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForEnvelopeStorage.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForEnvelopeStorage.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForEnvelopeStorage.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForEnvelopeStorage.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("EnvelopeStorageHandler: EnvelopeStorage handler failed to publish the response: %s", err)
			serverRequestsForEnvelopeStorage.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForEnvelopeStorage.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type EnvelopeStorageClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewEnvelopeStorageClient(nc nrpc.NatsConn) *EnvelopeStorageClient {
	return &EnvelopeStorageClient{
		nc:      nc,
		PkgSubject: "xbus.default",
		Subject: "envelopestorage",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *EnvelopeStorageClient) Retrieve(
	ctx context.Context,clientKind string,clientID string,
	req EnvelopePosition,
	cb func (context.Context, EnvelopeFragmentAndPosition),
) error {
	start := time.Now()
	subject := c.PkgSubject + "." + c.Subject + "." + "retrieve" + "." + clientKind + "." + clientID

	sub, err := nrpc.StreamCall(ctx, c.nc, subject, &req, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForEnvelopeStorage.WithLabelValues(
			"Retrieve", c.Encoding, "error").Inc()
		return err
	}

	var res EnvelopeFragmentAndPosition
	for {
		err = sub.Next(&res)
		if err != nil {
			break
		}
		cb(ctx, res)
	}
	if err == nrpc.ErrEOS {
		err = nil
	}
	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForEnvelopeStorage.WithLabelValues("Retrieve").Observe(elapsed)
	clientCallsForEnvelopeStorage.WithLabelValues(
		"Retrieve", c.Encoding, "success").Inc()
	return err
}

// ClientAPIServer is the interface that providers of the service
// ClientAPI should implement.
type ClientAPIServer interface {
	GetActors(ctx context.Context) (resp ClientGetActorsReply, err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForClientAPI = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "ClientAPI",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForClientAPI = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "ClientAPI",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForClientAPI = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "ClientAPI",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForClientAPI = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "ClientAPI",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// ClientAPIHandler provides a NATS subscription handler that can serve a
// subscription using a given ClientAPIServer implementation.
type ClientAPIHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  ClientAPIServer
}

func NewClientAPIHandler(ctx context.Context, nc nrpc.NatsConn, s ClientAPIServer) *ClientAPIHandler {
	return &ClientAPIHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewClientAPIConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s ClientAPIServer) *ClientAPIHandler {
	return &ClientAPIHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *ClientAPIHandler) Subject() string {
	return "xbus.default.client.*.>"
}

func (h *ClientAPIHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, svcParams, name, tail, err := nrpc.ParseSubject(
		"xbus.default", 0, "client", 1, msg.Subject)
	if err != nil {
		log.Printf("ClientAPIHanlder: ClientAPI subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail
	request.SetServiceParam("accountID", svcParams[0])

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "getactors":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("GetActorsHanlder: GetActors subject parsing failed: %v", err)
			break
		}
		var req github_com_nats_rpc_gogo_nrpc.Void
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("GetActorsHandler: GetActors request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForClientAPI.WithLabelValues(
				"GetActors", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.GetActors(ctx)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	default:
		log.Printf("ClientAPIHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForClientAPI.WithLabelValues(
			"ClientAPI", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForClientAPI.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForClientAPI.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForClientAPI.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForClientAPI.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("ClientAPIHandler: ClientAPI handler failed to publish the response: %s", err)
			serverRequestsForClientAPI.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForClientAPI.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type ClientAPIClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	SvcParamaccountID string
	Encoding string
	Timeout time.Duration
}

func NewClientAPIClient(nc nrpc.NatsConn, svcParamaccountID string) *ClientAPIClient {
	return &ClientAPIClient{
		nc:      nc,
		PkgSubject: "xbus.default",
		Subject: "client",
		SvcParamaccountID: svcParamaccountID,
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *ClientAPIClient) GetActors() (resp ClientGetActorsReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + c.SvcParamaccountID + "." + "getactors"

	// call
	var req github_com_nats_rpc_gogo_nrpc.Void
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForClientAPI.WithLabelValues(
			"GetActors", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForClientAPI.WithLabelValues("GetActors").Observe(elapsed)
	clientCallsForClientAPI.WithLabelValues(
		"GetActors", c.Encoding, "success").Inc()

	return
}

// WhoAmIServer is the interface that providers of the service
// WhoAmI should implement.
type WhoAmIServer interface {
	WhoAmI(ctx context.Context, identifier string) (resp Account, err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForWhoAmI = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "WhoAmI",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForWhoAmI = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "WhoAmI",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForWhoAmI = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "WhoAmI",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForWhoAmI = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "WhoAmI",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// WhoAmIHandler provides a NATS subscription handler that can serve a
// subscription using a given WhoAmIServer implementation.
type WhoAmIHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  WhoAmIServer
}

func NewWhoAmIHandler(ctx context.Context, nc nrpc.NatsConn, s WhoAmIServer) *WhoAmIHandler {
	return &WhoAmIHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewWhoAmIConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s WhoAmIServer) *WhoAmIHandler {
	return &WhoAmIHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *WhoAmIHandler) Subject() string {
	return "xbus.default.client.>"
}

func (h *WhoAmIHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.default", 0, "client", 0, msg.Subject)
	if err != nil {
		log.Printf("WhoAmIHanlder: WhoAmI subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "whoami":
		var mtParams []string
		mtParams, request.Encoding, err = nrpc.ParseSubjectTail(1, request.SubjectTail)
		if err != nil {
			log.Printf("WhoAmIHanlder: WhoAmI subject parsing failed: %v", err)
			break
		}
		var req github_com_nats_rpc_gogo_nrpc.Void
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("WhoAmIHandler: WhoAmI request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForWhoAmI.WithLabelValues(
				"WhoAmI", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.WhoAmI(ctx, mtParams[0])
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	default:
		log.Printf("WhoAmIHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForWhoAmI.WithLabelValues(
			"WhoAmI", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForWhoAmI.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForWhoAmI.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForWhoAmI.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForWhoAmI.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("WhoAmIHandler: WhoAmI handler failed to publish the response: %s", err)
			serverRequestsForWhoAmI.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForWhoAmI.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type WhoAmIClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewWhoAmIClient(nc nrpc.NatsConn) *WhoAmIClient {
	return &WhoAmIClient{
		nc:      nc,
		PkgSubject: "xbus.default",
		Subject: "client",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *WhoAmIClient) WhoAmI(identifier string, ) (resp Account, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "whoami" + "." + identifier

	// call
	var req github_com_nats_rpc_gogo_nrpc.Void
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForWhoAmI.WithLabelValues(
			"WhoAmI", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForWhoAmI.WithLabelValues("WhoAmI").Observe(elapsed)
	clientCallsForWhoAmI.WithLabelValues(
		"WhoAmI", c.Encoding, "success").Inc()

	return
}

// ProcessStateServer is the interface that providers of the service
// ProcessState should implement.
type ProcessStateServer interface {
}

var (
	// The request completion time, measured at client-side.
	clientRCTForProcessState = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "ProcessState",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForProcessState = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "ProcessState",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForProcessState = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "ProcessState",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForProcessState = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "ProcessState",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// ProcessStateHandler provides a NATS subscription handler that can serve a
// subscription using a given ProcessStateServer implementation.
type ProcessStateHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  ProcessStateServer
}

func NewProcessStateHandler(ctx context.Context, nc nrpc.NatsConn, s ProcessStateServer) *ProcessStateHandler {
	return &ProcessStateHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewProcessStateConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s ProcessStateServer) *ProcessStateHandler {
	return &ProcessStateHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *ProcessStateHandler) Subject() string {
	return "xbus.default.processstate.>"
}

func (h *ProcessStateHandler) EnvelopeStatesPublish(mtemitterID string, mtenvelopeID string, msg EmitterEnvelopeState) error {
	rawMsg, err := nrpc.Marshal("protobuf", &msg)
	if err != nil {
		log.Printf("ProcessStateHandler.EnvelopeStatesPublish: error marshaling the message: %s", err)
		return err
	}
	subject := "xbus.default."+ "processstate."+ "envelopestates" + "." + mtemitterID + "." + mtenvelopeID
	return h.nc.Publish(subject, rawMsg)
}

type ProcessStateClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewProcessStateClient(nc nrpc.NatsConn) *ProcessStateClient {
	return &ProcessStateClient{
		nc:      nc,
		PkgSubject: "xbus.default",
		Subject: "processstate",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *ProcessStateClient) EnvelopeStatesSubject(
	mtemitterID string,mtenvelopeID string,
) string {
	return c.PkgSubject + "." + c.Subject + "." + "envelopestates" + "." + mtemitterID + "." + mtenvelopeID
}

type ProcessStateEnvelopeStatesSubscription struct {
	*nats.Subscription
}

func (s *ProcessStateEnvelopeStatesSubscription) Next(timeout time.Duration) (next EmitterEnvelopeState, err error) {
	msg, err := s.Subscription.NextMsg(timeout)
	if err != nil {
		return
	}
	err = nrpc.Unmarshal("protobuf", msg.Data, &next)
	return
}

func (c *ProcessStateClient) EnvelopeStatesSubscribeSync(
	mtemitterID string,mtenvelopeID string,
) (sub *ProcessStateEnvelopeStatesSubscription, err error) {
	subject := c.EnvelopeStatesSubject(
		mtemitterID,mtenvelopeID,
	)
	natsSub, err := c.nc.SubscribeSync(subject)
	if err != nil {
		return
	}
	sub = &ProcessStateEnvelopeStatesSubscription{natsSub}
	return
}

func (c *ProcessStateClient) EnvelopeStatesSubscribe(
	mtemitterID string,mtenvelopeID string,
	handler func (EmitterEnvelopeState),
) (sub *nats.Subscription, err error) {
	subject := c.EnvelopeStatesSubject(
		mtemitterID,mtenvelopeID,
	)
	sub, err = c.nc.Subscribe(subject, func(msg *nats.Msg){
		var pmsg EmitterEnvelopeState
		err := nrpc.Unmarshal("protobuf", msg.Data, &pmsg)
		if err != nil {
			log.Printf("ProcessStateClient.EnvelopeStatesSubscribe: Error decoding, %s", err)
			return
		}
		handler(pmsg)
	})
	return
}

func (c *ProcessStateClient) EnvelopeStatesSubscribeChan(
	mtemitterID string,mtenvelopeID string,
) (<-chan EmitterEnvelopeState, *nats.Subscription, error) {
	ch := make(chan EmitterEnvelopeState)
	sub, err := c.EnvelopeStatesSubscribe(mtemitterID, mtenvelopeID, func (msg EmitterEnvelopeState) {
		ch <- msg
	})
	return ch, sub, err
}

// ActorAgentServer is the interface that providers of the service
// ActorAgent should implement.
type ActorAgentServer interface {
	Process(ctx context.Context, req ActorProcessRequest, pushRep func(ActorProcessingState)) (err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForActorAgent = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "ActorAgent",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForActorAgent = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "ActorAgent",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForActorAgent = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "ActorAgent",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForActorAgent = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "ActorAgent",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// ActorAgentHandler provides a NATS subscription handler that can serve a
// subscription using a given ActorAgentServer implementation.
type ActorAgentHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  ActorAgentServer
}

func NewActorAgentHandler(ctx context.Context, nc nrpc.NatsConn, s ActorAgentServer) *ActorAgentHandler {
	return &ActorAgentHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewActorAgentConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s ActorAgentServer) *ActorAgentHandler {
	return &ActorAgentHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *ActorAgentHandler) Subject() string {
	return "xbus.default.actoragent.*.>"
}

func (h *ActorAgentHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, svcParams, name, tail, err := nrpc.ParseSubject(
		"xbus.default", 0, "actoragent", 1, msg.Subject)
	if err != nil {
		log.Printf("ActorAgentHanlder: ActorAgent subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail
	request.SetServiceParam("actorID", svcParams[0])

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "process":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("ProcessHanlder: Process subject parsing failed: %v", err)
			break
		}
		var req ActorProcessRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("ProcessHandler: Process request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForActorAgent.WithLabelValues(
				"Process", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.EnableStreamedReply()
			request.Handler = func(ctx context.Context)(proto.Message, error){
				err := h.server.Process(ctx, req, func(rep ActorProcessingState){
					request.SendStreamReply(&rep)
				})
				return nil, err
			}
		}
	default:
		log.Printf("ActorAgentHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForActorAgent.WithLabelValues(
			"ActorAgent", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForActorAgent.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForActorAgent.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForActorAgent.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForActorAgent.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("ActorAgentHandler: ActorAgent handler failed to publish the response: %s", err)
			serverRequestsForActorAgent.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForActorAgent.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type ActorAgentClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	SvcParamactorID string
	Encoding string
	Timeout time.Duration
}

func NewActorAgentClient(nc nrpc.NatsConn, svcParamactorID string) *ActorAgentClient {
	return &ActorAgentClient{
		nc:      nc,
		PkgSubject: "xbus.default",
		Subject: "actoragent",
		SvcParamactorID: svcParamactorID,
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *ActorAgentClient) Process(
	ctx context.Context,
	req ActorProcessRequest,
	cb func (context.Context, ActorProcessingState),
) error {
	start := time.Now()
	subject := c.PkgSubject + "." + c.Subject + "." + c.SvcParamactorID + "." + "process"

	sub, err := nrpc.StreamCall(ctx, c.nc, subject, &req, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForActorAgent.WithLabelValues(
			"Process", c.Encoding, "error").Inc()
		return err
	}

	var res ActorProcessingState
	for {
		err = sub.Next(&res)
		if err != nil {
			break
		}
		cb(ctx, res)
	}
	if err == nrpc.ErrEOS {
		err = nil
	}
	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForActorAgent.WithLabelValues("Process").Observe(elapsed)
	clientCallsForActorAgent.WithLabelValues(
		"Process", c.Encoding, "success").Inc()
	return err
}

// BrokerServer is the interface that providers of the service
// Broker should implement.
type BrokerServer interface {
	Output(ctx context.Context, req OutputRequest) (resp EnvelopeAck, err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForBroker = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Broker",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForBroker = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Broker",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForBroker = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "Broker",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForBroker = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "Broker",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// BrokerHandler provides a NATS subscription handler that can serve a
// subscription using a given BrokerServer implementation.
type BrokerHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  BrokerServer
}

func NewBrokerHandler(ctx context.Context, nc nrpc.NatsConn, s BrokerServer) *BrokerHandler {
	return &BrokerHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewBrokerConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s BrokerServer) *BrokerHandler {
	return &BrokerHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *BrokerHandler) Subject() string {
	return "xbus.default.broker.*.>"
}

func (h *BrokerHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, svcParams, name, tail, err := nrpc.ParseSubject(
		"xbus.default", 0, "broker", 1, msg.Subject)
	if err != nil {
		log.Printf("BrokerHanlder: Broker subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail
	request.SetServiceParam("actorID", svcParams[0])

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "output":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("OutputHanlder: Output subject parsing failed: %v", err)
			break
		}
		var req OutputRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("OutputHandler: Output request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForBroker.WithLabelValues(
				"Output", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Output(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	default:
		log.Printf("BrokerHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForBroker.WithLabelValues(
			"Broker", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForBroker.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForBroker.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForBroker.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForBroker.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("BrokerHandler: Broker handler failed to publish the response: %s", err)
			serverRequestsForBroker.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForBroker.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type BrokerClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	SvcParamactorID string
	Encoding string
	Timeout time.Duration
}

func NewBrokerClient(nc nrpc.NatsConn, svcParamactorID string) *BrokerClient {
	return &BrokerClient{
		nc:      nc,
		PkgSubject: "xbus.default",
		Subject: "broker",
		SvcParamactorID: svcParamactorID,
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *BrokerClient) Output(req OutputRequest) (resp EnvelopeAck, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + c.SvcParamactorID + "." + "output"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForBroker.WithLabelValues(
			"Output", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForBroker.WithLabelValues("Output").Observe(elapsed)
	clientCallsForBroker.WithLabelValues(
		"Output", c.Encoding, "success").Inc()

	return
}

// CoreEventsServer is the interface that providers of the service
// CoreEvents should implement.
type CoreEventsServer interface {
}

var (
	// The request completion time, measured at client-side.
	clientRCTForCoreEvents = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "CoreEvents",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForCoreEvents = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "CoreEvents",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForCoreEvents = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "CoreEvents",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForCoreEvents = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "CoreEvents",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// CoreEventsHandler provides a NATS subscription handler that can serve a
// subscription using a given CoreEventsServer implementation.
type CoreEventsHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  CoreEventsServer
}

func NewCoreEventsHandler(ctx context.Context, nc nrpc.NatsConn, s CoreEventsServer) *CoreEventsHandler {
	return &CoreEventsHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewCoreEventsConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s CoreEventsServer) *CoreEventsHandler {
	return &CoreEventsHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *CoreEventsHandler) Subject() string {
	return "xbus.default.core.events.>"
}

func (h *CoreEventsHandler) ActorPublish(msg CoreEvent) error {
	rawMsg, err := nrpc.Marshal("protobuf", &msg)
	if err != nil {
		log.Printf("CoreEventsHandler.ActorPublish: error marshaling the message: %s", err)
		return err
	}
	subject := "xbus.default."+ "core.events."+ "actor"
	return h.nc.Publish(subject, rawMsg)
}

func (h *CoreEventsHandler) ProcessPublish(msg CoreEvent) error {
	rawMsg, err := nrpc.Marshal("protobuf", &msg)
	if err != nil {
		log.Printf("CoreEventsHandler.ProcessPublish: error marshaling the message: %s", err)
		return err
	}
	subject := "xbus.default."+ "core.events."+ "process"
	return h.nc.Publish(subject, rawMsg)
}

func (h *CoreEventsHandler) ProcessNodeReadyPublish(msg ProcessNodeReadyEvent) error {
	rawMsg, err := nrpc.Marshal("protobuf", &msg)
	if err != nil {
		log.Printf("CoreEventsHandler.ProcessNodeReadyPublish: error marshaling the message: %s", err)
		return err
	}
	subject := "xbus.default."+ "core.events."+ "processnodeready"
	return h.nc.Publish(subject, rawMsg)
}

func (h *CoreEventsHandler) EnvelopePublish(msg EnvelopeEvent) error {
	rawMsg, err := nrpc.Marshal("protobuf", &msg)
	if err != nil {
		log.Printf("CoreEventsHandler.EnvelopePublish: error marshaling the message: %s", err)
		return err
	}
	subject := "xbus.default."+ "core.events."+ "envelope"
	return h.nc.Publish(subject, rawMsg)
}

type CoreEventsClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewCoreEventsClient(nc nrpc.NatsConn) *CoreEventsClient {
	return &CoreEventsClient{
		nc:      nc,
		PkgSubject: "xbus.default",
		Subject: "core.events",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *CoreEventsClient) ActorSubject(
	
) string {
	return c.PkgSubject + "." + c.Subject + "." + "actor"
}

type CoreEventsActorSubscription struct {
	*nats.Subscription
}

func (s *CoreEventsActorSubscription) Next(timeout time.Duration) (next CoreEvent, err error) {
	msg, err := s.Subscription.NextMsg(timeout)
	if err != nil {
		return
	}
	err = nrpc.Unmarshal("protobuf", msg.Data, &next)
	return
}

func (c *CoreEventsClient) ActorSubscribeSync(
	
) (sub *CoreEventsActorSubscription, err error) {
	subject := c.ActorSubject(
		
	)
	natsSub, err := c.nc.SubscribeSync(subject)
	if err != nil {
		return
	}
	sub = &CoreEventsActorSubscription{natsSub}
	return
}

func (c *CoreEventsClient) ActorSubscribe(
	
	handler func (CoreEvent),
) (sub *nats.Subscription, err error) {
	subject := c.ActorSubject(
		
	)
	sub, err = c.nc.Subscribe(subject, func(msg *nats.Msg){
		var pmsg CoreEvent
		err := nrpc.Unmarshal("protobuf", msg.Data, &pmsg)
		if err != nil {
			log.Printf("CoreEventsClient.ActorSubscribe: Error decoding, %s", err)
			return
		}
		handler(pmsg)
	})
	return
}

func (c *CoreEventsClient) ActorSubscribeChan(
	
) (<-chan CoreEvent, *nats.Subscription, error) {
	ch := make(chan CoreEvent)
	sub, err := c.ActorSubscribe(func (msg CoreEvent) {
		ch <- msg
	})
	return ch, sub, err
}

func (c *CoreEventsClient) ProcessSubject(
	
) string {
	return c.PkgSubject + "." + c.Subject + "." + "process"
}

type CoreEventsProcessSubscription struct {
	*nats.Subscription
}

func (s *CoreEventsProcessSubscription) Next(timeout time.Duration) (next CoreEvent, err error) {
	msg, err := s.Subscription.NextMsg(timeout)
	if err != nil {
		return
	}
	err = nrpc.Unmarshal("protobuf", msg.Data, &next)
	return
}

func (c *CoreEventsClient) ProcessSubscribeSync(
	
) (sub *CoreEventsProcessSubscription, err error) {
	subject := c.ProcessSubject(
		
	)
	natsSub, err := c.nc.SubscribeSync(subject)
	if err != nil {
		return
	}
	sub = &CoreEventsProcessSubscription{natsSub}
	return
}

func (c *CoreEventsClient) ProcessSubscribe(
	
	handler func (CoreEvent),
) (sub *nats.Subscription, err error) {
	subject := c.ProcessSubject(
		
	)
	sub, err = c.nc.Subscribe(subject, func(msg *nats.Msg){
		var pmsg CoreEvent
		err := nrpc.Unmarshal("protobuf", msg.Data, &pmsg)
		if err != nil {
			log.Printf("CoreEventsClient.ProcessSubscribe: Error decoding, %s", err)
			return
		}
		handler(pmsg)
	})
	return
}

func (c *CoreEventsClient) ProcessSubscribeChan(
	
) (<-chan CoreEvent, *nats.Subscription, error) {
	ch := make(chan CoreEvent)
	sub, err := c.ProcessSubscribe(func (msg CoreEvent) {
		ch <- msg
	})
	return ch, sub, err
}

func (c *CoreEventsClient) ProcessNodeReadySubject(
	
) string {
	return c.PkgSubject + "." + c.Subject + "." + "processnodeready"
}

type CoreEventsProcessNodeReadySubscription struct {
	*nats.Subscription
}

func (s *CoreEventsProcessNodeReadySubscription) Next(timeout time.Duration) (next ProcessNodeReadyEvent, err error) {
	msg, err := s.Subscription.NextMsg(timeout)
	if err != nil {
		return
	}
	err = nrpc.Unmarshal("protobuf", msg.Data, &next)
	return
}

func (c *CoreEventsClient) ProcessNodeReadySubscribeSync(
	
) (sub *CoreEventsProcessNodeReadySubscription, err error) {
	subject := c.ProcessNodeReadySubject(
		
	)
	natsSub, err := c.nc.SubscribeSync(subject)
	if err != nil {
		return
	}
	sub = &CoreEventsProcessNodeReadySubscription{natsSub}
	return
}

func (c *CoreEventsClient) ProcessNodeReadySubscribe(
	
	handler func (ProcessNodeReadyEvent),
) (sub *nats.Subscription, err error) {
	subject := c.ProcessNodeReadySubject(
		
	)
	sub, err = c.nc.Subscribe(subject, func(msg *nats.Msg){
		var pmsg ProcessNodeReadyEvent
		err := nrpc.Unmarshal("protobuf", msg.Data, &pmsg)
		if err != nil {
			log.Printf("CoreEventsClient.ProcessNodeReadySubscribe: Error decoding, %s", err)
			return
		}
		handler(pmsg)
	})
	return
}

func (c *CoreEventsClient) ProcessNodeReadySubscribeChan(
	
) (<-chan ProcessNodeReadyEvent, *nats.Subscription, error) {
	ch := make(chan ProcessNodeReadyEvent)
	sub, err := c.ProcessNodeReadySubscribe(func (msg ProcessNodeReadyEvent) {
		ch <- msg
	})
	return ch, sub, err
}

func (c *CoreEventsClient) EnvelopeSubject(
	
) string {
	return c.PkgSubject + "." + c.Subject + "." + "envelope"
}

type CoreEventsEnvelopeSubscription struct {
	*nats.Subscription
}

func (s *CoreEventsEnvelopeSubscription) Next(timeout time.Duration) (next EnvelopeEvent, err error) {
	msg, err := s.Subscription.NextMsg(timeout)
	if err != nil {
		return
	}
	err = nrpc.Unmarshal("protobuf", msg.Data, &next)
	return
}

func (c *CoreEventsClient) EnvelopeSubscribeSync(
	
) (sub *CoreEventsEnvelopeSubscription, err error) {
	subject := c.EnvelopeSubject(
		
	)
	natsSub, err := c.nc.SubscribeSync(subject)
	if err != nil {
		return
	}
	sub = &CoreEventsEnvelopeSubscription{natsSub}
	return
}

func (c *CoreEventsClient) EnvelopeSubscribe(
	
	handler func (EnvelopeEvent),
) (sub *nats.Subscription, err error) {
	subject := c.EnvelopeSubject(
		
	)
	sub, err = c.nc.Subscribe(subject, func(msg *nats.Msg){
		var pmsg EnvelopeEvent
		err := nrpc.Unmarshal("protobuf", msg.Data, &pmsg)
		if err != nil {
			log.Printf("CoreEventsClient.EnvelopeSubscribe: Error decoding, %s", err)
			return
		}
		handler(pmsg)
	})
	return
}

func (c *CoreEventsClient) EnvelopeSubscribeChan(
	
) (<-chan EnvelopeEvent, *nats.Subscription, error) {
	ch := make(chan EnvelopeEvent)
	sub, err := c.EnvelopeSubscribe(func (msg EnvelopeEvent) {
		ch <- msg
	})
	return ch, sub, err
}

// SvcRegistrationServer is the interface that providers of the service
// SvcRegistration should implement.
type SvcRegistrationServer interface {
	Register(ctx context.Context, req RegistrationRequest) (resp Registration, err error)
	Update(ctx context.Context, accountID string, req RegistrationRequest) (resp Registration, err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForSvcRegistration = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "SvcRegistration",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForSvcRegistration = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "SvcRegistration",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForSvcRegistration = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "SvcRegistration",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForSvcRegistration = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "SvcRegistration",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// SvcRegistrationHandler provides a NATS subscription handler that can serve a
// subscription using a given SvcRegistrationServer implementation.
type SvcRegistrationHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  SvcRegistrationServer
}

func NewSvcRegistrationHandler(ctx context.Context, nc nrpc.NatsConn, s SvcRegistrationServer) *SvcRegistrationHandler {
	return &SvcRegistrationHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewSvcRegistrationConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s SvcRegistrationServer) *SvcRegistrationHandler {
	return &SvcRegistrationHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *SvcRegistrationHandler) Subject() string {
	return "xbus.default.registration.>"
}

func (h *SvcRegistrationHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.default", 0, "registration", 0, msg.Subject)
	if err != nil {
		log.Printf("SvcRegistrationHanlder: SvcRegistration subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "register":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("RegisterHanlder: Register subject parsing failed: %v", err)
			break
		}
		var req RegistrationRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("RegisterHandler: Register request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForSvcRegistration.WithLabelValues(
				"Register", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Register(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "update":
		var mtParams []string
		mtParams, request.Encoding, err = nrpc.ParseSubjectTail(1, request.SubjectTail)
		if err != nil {
			log.Printf("UpdateHanlder: Update subject parsing failed: %v", err)
			break
		}
		var req RegistrationRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("UpdateHandler: Update request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForSvcRegistration.WithLabelValues(
				"Update", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Update(ctx, mtParams[0], req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	default:
		log.Printf("SvcRegistrationHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForSvcRegistration.WithLabelValues(
			"SvcRegistration", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForSvcRegistration.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForSvcRegistration.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForSvcRegistration.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForSvcRegistration.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("SvcRegistrationHandler: SvcRegistration handler failed to publish the response: %s", err)
			serverRequestsForSvcRegistration.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForSvcRegistration.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type SvcRegistrationClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewSvcRegistrationClient(nc nrpc.NatsConn) *SvcRegistrationClient {
	return &SvcRegistrationClient{
		nc:      nc,
		PkgSubject: "xbus.default",
		Subject: "registration",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *SvcRegistrationClient) Register(req RegistrationRequest) (resp Registration, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "register"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForSvcRegistration.WithLabelValues(
			"Register", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForSvcRegistration.WithLabelValues("Register").Observe(elapsed)
	clientCallsForSvcRegistration.WithLabelValues(
		"Register", c.Encoding, "success").Inc()

	return
}

func (c *SvcRegistrationClient) Update(accountID string, req RegistrationRequest) (resp Registration, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "update" + "." + accountID

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForSvcRegistration.WithLabelValues(
			"Update", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForSvcRegistration.WithLabelValues("Update").Observe(elapsed)
	clientCallsForSvcRegistration.WithLabelValues(
		"Update", c.Encoding, "success").Inc()

	return
}

// SentinelServer is the interface that providers of the service
// Sentinel should implement.
type SentinelServer interface {
	ActorReady(ctx context.Context, actorID string)
	ActorLeaving(ctx context.Context, actorID string)
	ListActors(ctx context.Context, pushRep func(ClientGetActorsReply)) (err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForSentinel = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Sentinel",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForSentinel = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Sentinel",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForSentinel = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "Sentinel",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForSentinel = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "Sentinel",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// SentinelHandler provides a NATS subscription handler that can serve a
// subscription using a given SentinelServer implementation.
type SentinelHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  SentinelServer
}

func NewSentinelHandler(ctx context.Context, nc nrpc.NatsConn, s SentinelServer) *SentinelHandler {
	return &SentinelHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewSentinelConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s SentinelServer) *SentinelHandler {
	return &SentinelHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *SentinelHandler) Subject() string {
	return "xbus.default.sentinel.>"
}

func (h *SentinelHandler) SurveyPublish(msg SurveyRequest) error {
	rawMsg, err := nrpc.Marshal("protobuf", &msg)
	if err != nil {
		log.Printf("SentinelHandler.SurveyPublish: error marshaling the message: %s", err)
		return err
	}
	subject := "xbus.default."+ "sentinel."+ "survey"
	return h.nc.Publish(subject, rawMsg)
}

func (h *SentinelHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.default", 0, "sentinel", 0, msg.Subject)
	if err != nil {
		log.Printf("SentinelHanlder: Sentinel subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "survey":
		// Survey is a no-request method. Ignore it.
		return
	case "actorready":
		var mtParams []string
		request.NoReply = true
		mtParams, request.Encoding, err = nrpc.ParseSubjectTail(1, request.SubjectTail)
		if err != nil {
			log.Printf("ActorReadyHanlder: ActorReady subject parsing failed: %v", err)
			break
		}
		var req github_com_nats_rpc_gogo_nrpc.Void
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("ActorReadyHandler: ActorReady request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForSentinel.WithLabelValues(
				"ActorReady", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){var innerResp nrpc.NoReply
				h.server.ActorReady(ctx, mtParams[0])
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "actorleaving":
		var mtParams []string
		request.NoReply = true
		mtParams, request.Encoding, err = nrpc.ParseSubjectTail(1, request.SubjectTail)
		if err != nil {
			log.Printf("ActorLeavingHanlder: ActorLeaving subject parsing failed: %v", err)
			break
		}
		var req github_com_nats_rpc_gogo_nrpc.Void
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("ActorLeavingHandler: ActorLeaving request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForSentinel.WithLabelValues(
				"ActorLeaving", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){var innerResp nrpc.NoReply
				h.server.ActorLeaving(ctx, mtParams[0])
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "listactors":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("ListActorsHanlder: ListActors subject parsing failed: %v", err)
			break
		}
		var req github_com_nats_rpc_gogo_nrpc.Void
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("ListActorsHandler: ListActors request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForSentinel.WithLabelValues(
				"ListActors", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.EnableStreamedReply()
			request.Handler = func(ctx context.Context)(proto.Message, error){
				err := h.server.ListActors(ctx, func(rep ClientGetActorsReply){
					request.SendStreamReply(&rep)
				})
				return nil, err
			}
		}
	default:
		log.Printf("SentinelHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForSentinel.WithLabelValues(
			"Sentinel", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForSentinel.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForSentinel.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForSentinel.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForSentinel.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("SentinelHandler: Sentinel handler failed to publish the response: %s", err)
			serverRequestsForSentinel.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForSentinel.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type SentinelClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewSentinelClient(nc nrpc.NatsConn) *SentinelClient {
	return &SentinelClient{
		nc:      nc,
		PkgSubject: "xbus.default",
		Subject: "sentinel",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *SentinelClient) SurveySubject(
	
) string {
	return c.PkgSubject + "." + c.Subject + "." + "survey"
}

type SentinelSurveySubscription struct {
	*nats.Subscription
}

func (s *SentinelSurveySubscription) Next(timeout time.Duration) (next SurveyRequest, err error) {
	msg, err := s.Subscription.NextMsg(timeout)
	if err != nil {
		return
	}
	err = nrpc.Unmarshal("protobuf", msg.Data, &next)
	return
}

func (c *SentinelClient) SurveySubscribeSync(
	
) (sub *SentinelSurveySubscription, err error) {
	subject := c.SurveySubject(
		
	)
	natsSub, err := c.nc.SubscribeSync(subject)
	if err != nil {
		return
	}
	sub = &SentinelSurveySubscription{natsSub}
	return
}

func (c *SentinelClient) SurveySubscribe(
	
	handler func (SurveyRequest),
) (sub *nats.Subscription, err error) {
	subject := c.SurveySubject(
		
	)
	sub, err = c.nc.Subscribe(subject, func(msg *nats.Msg){
		var pmsg SurveyRequest
		err := nrpc.Unmarshal("protobuf", msg.Data, &pmsg)
		if err != nil {
			log.Printf("SentinelClient.SurveySubscribe: Error decoding, %s", err)
			return
		}
		handler(pmsg)
	})
	return
}

func (c *SentinelClient) SurveySubscribeChan(
	
) (<-chan SurveyRequest, *nats.Subscription, error) {
	ch := make(chan SurveyRequest)
	sub, err := c.SurveySubscribe(func (msg SurveyRequest) {
		ch <- msg
	})
	return ch, sub, err
}

func (c *SentinelClient) ActorReady(actorID string, ) (err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "actorready" + "." + actorID

	// call
	var req github_com_nats_rpc_gogo_nrpc.Void
	var resp github_com_nats_rpc_gogo_nrpc.NoReply
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForSentinel.WithLabelValues(
			"ActorReady", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForSentinel.WithLabelValues("ActorReady").Observe(elapsed)
	clientCallsForSentinel.WithLabelValues(
		"ActorReady", c.Encoding, "success").Inc()

	return
}

func (c *SentinelClient) ActorLeaving(actorID string, ) (err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "actorleaving" + "." + actorID

	// call
	var req github_com_nats_rpc_gogo_nrpc.Void
	var resp github_com_nats_rpc_gogo_nrpc.NoReply
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForSentinel.WithLabelValues(
			"ActorLeaving", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForSentinel.WithLabelValues("ActorLeaving").Observe(elapsed)
	clientCallsForSentinel.WithLabelValues(
		"ActorLeaving", c.Encoding, "success").Inc()

	return
}

func (c *SentinelClient) ListActors(
	ctx context.Context,
	cb func (context.Context, ClientGetActorsReply),
) error {
	start := time.Now()
	subject := c.PkgSubject + "." + c.Subject + "." + "listactors"

	sub, err := nrpc.StreamCall(ctx, c.nc, subject, &nrpc.Void{}, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForSentinel.WithLabelValues(
			"ListActors", c.Encoding, "error").Inc()
		return err
	}

	var res ClientGetActorsReply
	for {
		err = sub.Next(&res)
		if err != nil {
			break
		}
		cb(ctx, res)
	}
	if err == nrpc.ErrEOS {
		err = nil
	}
	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForSentinel.WithLabelValues("ListActors").Observe(elapsed)
	clientCallsForSentinel.WithLabelValues(
		"ListActors", c.Encoding, "success").Inc()
	return err
}

type Client struct {
	nc      nrpc.NatsConn
	defaultEncoding string
	defaultTimeout time.Duration
	pkgSubject string
	Director *DirectorClient
	EnvelopeStorage *EnvelopeStorageClient
	ClientAPI *ClientAPIClient
	WhoAmI *WhoAmIClient
	ProcessState *ProcessStateClient
	ActorAgent *ActorAgentClient
	Broker *BrokerClient
	CoreEvents *CoreEventsClient
	SvcRegistration *SvcRegistrationClient
	Sentinel *SentinelClient
}

func NewClient(nc nrpc.NatsConn) *Client {
	c := Client{
		nc: nc,
		defaultEncoding: "protobuf",
		defaultTimeout: 5*time.Second,
		pkgSubject: "xbus.default",
	}
	c.Director = NewDirectorClient(nc)
	c.EnvelopeStorage = NewEnvelopeStorageClient(nc)
	c.WhoAmI = NewWhoAmIClient(nc)
	c.ProcessState = NewProcessStateClient(nc)
	c.CoreEvents = NewCoreEventsClient(nc)
	c.SvcRegistration = NewSvcRegistrationClient(nc)
	c.Sentinel = NewSentinelClient(nc)
	return &c
}

func (c *Client) SetEncoding(encoding string) {
	c.defaultEncoding = encoding
	if c.Director != nil {
		c.Director.Encoding = encoding
	}
	if c.EnvelopeStorage != nil {
		c.EnvelopeStorage.Encoding = encoding
	}
	if c.ClientAPI != nil {
		c.ClientAPI.Encoding = encoding
	}
	if c.WhoAmI != nil {
		c.WhoAmI.Encoding = encoding
	}
	if c.ProcessState != nil {
		c.ProcessState.Encoding = encoding
	}
	if c.ActorAgent != nil {
		c.ActorAgent.Encoding = encoding
	}
	if c.Broker != nil {
		c.Broker.Encoding = encoding
	}
	if c.CoreEvents != nil {
		c.CoreEvents.Encoding = encoding
	}
	if c.SvcRegistration != nil {
		c.SvcRegistration.Encoding = encoding
	}
	if c.Sentinel != nil {
		c.Sentinel.Encoding = encoding
	}
}

func (c *Client) SetTimeout(t time.Duration) {
	c.defaultTimeout = t
	if c.Director != nil {
		c.Director.Timeout = t
	}
	if c.EnvelopeStorage != nil {
		c.EnvelopeStorage.Timeout = t
	}
	if c.ClientAPI != nil {
		c.ClientAPI.Timeout = t
	}
	if c.WhoAmI != nil {
		c.WhoAmI.Timeout = t
	}
	if c.ProcessState != nil {
		c.ProcessState.Timeout = t
	}
	if c.ActorAgent != nil {
		c.ActorAgent.Timeout = t
	}
	if c.Broker != nil {
		c.Broker.Timeout = t
	}
	if c.CoreEvents != nil {
		c.CoreEvents.Timeout = t
	}
	if c.SvcRegistration != nil {
		c.SvcRegistration.Timeout = t
	}
	if c.Sentinel != nil {
		c.Sentinel.Timeout = t
	}
}

func (c *Client) SetClientAPIParams(
	accountID string,
) {
	c.ClientAPI = NewClientAPIClient(
		c.nc,
		accountID,
	)
	c.ClientAPI.Encoding = c.defaultEncoding
	c.ClientAPI.Timeout = c.defaultTimeout
}

func (c *Client) NewClientAPI(
	accountID string,
) *ClientAPIClient {
	client := NewClientAPIClient(
		c.nc,
		accountID,
	)
	client.Encoding = c.defaultEncoding
	client.Timeout = c.defaultTimeout
	return client
}

func (c *Client) SetActorAgentParams(
	actorID string,
) {
	c.ActorAgent = NewActorAgentClient(
		c.nc,
		actorID,
	)
	c.ActorAgent.Encoding = c.defaultEncoding
	c.ActorAgent.Timeout = c.defaultTimeout
}

func (c *Client) NewActorAgent(
	actorID string,
) *ActorAgentClient {
	client := NewActorAgentClient(
		c.nc,
		actorID,
	)
	client.Encoding = c.defaultEncoding
	client.Timeout = c.defaultTimeout
	return client
}

func (c *Client) SetBrokerParams(
	actorID string,
) {
	c.Broker = NewBrokerClient(
		c.nc,
		actorID,
	)
	c.Broker.Encoding = c.defaultEncoding
	c.Broker.Timeout = c.defaultTimeout
}

func (c *Client) NewBroker(
	actorID string,
) *BrokerClient {
	client := NewBrokerClient(
		c.nc,
		actorID,
	)
	client.Encoding = c.defaultEncoding
	client.Timeout = c.defaultTimeout
	return client
}

func init() {
	// register metrics for service Director
	prometheus.MustRegister(clientRCTForDirector)
	prometheus.MustRegister(serverHETForDirector)
	prometheus.MustRegister(clientCallsForDirector)
	prometheus.MustRegister(serverRequestsForDirector)
	// register metrics for service EnvelopeStorage
	prometheus.MustRegister(clientRCTForEnvelopeStorage)
	prometheus.MustRegister(serverHETForEnvelopeStorage)
	prometheus.MustRegister(clientCallsForEnvelopeStorage)
	prometheus.MustRegister(serverRequestsForEnvelopeStorage)
	// register metrics for service ClientAPI
	prometheus.MustRegister(clientRCTForClientAPI)
	prometheus.MustRegister(serverHETForClientAPI)
	prometheus.MustRegister(clientCallsForClientAPI)
	prometheus.MustRegister(serverRequestsForClientAPI)
	// register metrics for service WhoAmI
	prometheus.MustRegister(clientRCTForWhoAmI)
	prometheus.MustRegister(serverHETForWhoAmI)
	prometheus.MustRegister(clientCallsForWhoAmI)
	prometheus.MustRegister(serverRequestsForWhoAmI)
	// register metrics for service ProcessState
	prometheus.MustRegister(clientRCTForProcessState)
	prometheus.MustRegister(serverHETForProcessState)
	prometheus.MustRegister(clientCallsForProcessState)
	prometheus.MustRegister(serverRequestsForProcessState)
	// register metrics for service ActorAgent
	prometheus.MustRegister(clientRCTForActorAgent)
	prometheus.MustRegister(serverHETForActorAgent)
	prometheus.MustRegister(clientCallsForActorAgent)
	prometheus.MustRegister(serverRequestsForActorAgent)
	// register metrics for service Broker
	prometheus.MustRegister(clientRCTForBroker)
	prometheus.MustRegister(serverHETForBroker)
	prometheus.MustRegister(clientCallsForBroker)
	prometheus.MustRegister(serverRequestsForBroker)
	// register metrics for service CoreEvents
	prometheus.MustRegister(clientRCTForCoreEvents)
	prometheus.MustRegister(serverHETForCoreEvents)
	prometheus.MustRegister(clientCallsForCoreEvents)
	prometheus.MustRegister(serverRequestsForCoreEvents)
	// register metrics for service SvcRegistration
	prometheus.MustRegister(clientRCTForSvcRegistration)
	prometheus.MustRegister(serverHETForSvcRegistration)
	prometheus.MustRegister(clientCallsForSvcRegistration)
	prometheus.MustRegister(serverRequestsForSvcRegistration)
	// register metrics for service Sentinel
	prometheus.MustRegister(clientRCTForSentinel)
	prometheus.MustRegister(serverHETForSentinel)
	prometheus.MustRegister(clientCallsForSentinel)
	prometheus.MustRegister(serverRequestsForSentinel)
}